/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.tools;

import java.io.ObjectStreamClass;
import java.io.Serializable;
import java.io.FileInputStream;
import java.io.ObjectInputStream;
import java.util.Map;

/**
 * Encapsulates a class serialVersionUID and codebase.
 * 
 * @author Scott.Stark@jboss.org
 * @version $Revision: 1.1.2.2 $
 */
public class ClassVersionInfo implements Serializable
{
   static final long serialVersionUID = 2036506209171911437L;

   /** The named class serialVersionUID as returned by ObjectStreamClass */
   private long serialVersion;
   /** The binary class name */
   private String name;
   private boolean hasExplicitSerialVersionUID;

   public ClassVersionInfo() {}

   public ClassVersionInfo(String name, ClassLoader loader)
      throws ClassNotFoundException
   {
      this.name = name;
      Class c = loader.loadClass(name);
      if( c.isInterface() == false )
      {
         ObjectStreamClass osc = ObjectStreamClass.lookup(c);
         if( osc != null )
         {
            serialVersion = osc.getSerialVersionUID();
            try
            {
               c.getDeclaredField("serialVersionUID");
               hasExplicitSerialVersionUID = true;
            }
            catch(NoSuchFieldException e)
            {
               hasExplicitSerialVersionUID = false;
            }
         }
      }
   }

   public long getSerialVersion()
   {
      return serialVersion;
   }

   public void setSerialVersion(long serialVersion)
   {
      this.serialVersion = serialVersion;
   }

   public boolean getHasExplicitSerialVersionUID()
   {
      return hasExplicitSerialVersionUID;
   }

   public void setHasExplicitSerialVersionUID(boolean flag)
   {
      hasExplicitSerialVersionUID = flag;
   }

   public String getName()
   {
      return name;
   }

   public void setName(String name)
   {
      this.name = name;
   }

   public String toString()
   {
      StringBuffer tmp = new StringBuffer("ClassVersionInfo");
      tmp.append('{');
      tmp.append("serialVersion=");
      tmp.append(serialVersion);
      tmp.append(", hasExplicitSerialVersionUID=");
      tmp.append(hasExplicitSerialVersionUID);
      tmp.append(", name=");
      tmp.append(name);
      tmp.append('}');
      return tmp.toString();
   }

   /**
    * Utility main entry point that allows one to load a Map<String,ClassVersionInfo>
    * from a serialized object file to print a specific ClassVersionInfo
    * 
    * @param args [0] = map.ser file for serialized object image
    *    [1] = class name of the ClassVersionInfo to print 
    * @throws Exception
    */
   public static void main(String[] args)
      throws Exception
   {
      if( args.length != 2 )
      {
         throw new IllegalArgumentException("Usage ClassVersionInfo map.ser [-list|cvi_fqn]");
      }
      FileInputStream fis = new FileInputStream(args[0]);
      ObjectInputStream ois = new ObjectInputStream(fis);
      Map cvis = (Map) ois.readObject();
      ois.close();

      if( args[1].equals("-list") )
      {
         System.out.println(cvis.keySet());
         System.exit(0);
      }

      String fqn = args[1];
      if( cvis.containsKey(fqn) == false )
      {
         System.err.println("No entry found for: "+fqn);
         System.exit(1);
      }
      ClassVersionInfo cvi = (ClassVersionInfo) cvis.get(fqn);
      System.out.println(cvi);
      System.exit(0);
   }
}
